@@ -1,80 +1,99 @@
-"""Deck class: serialization and card access under conditions"""
-from typing import Callable, Generator, Optional, TextIO
+"""Deck class: serialization and card access under conditions."""
+
+from typing import Any, Callable, Dict, Generator, Optional, TextIO
 import arrow
 import collections
 import datetime
 import email.utils
 import itertools
 import logging
+import re
 import requests
 import unidecode
 import urllib.parse
 
 from . import config
-from . import parser
 from . import vtes
 from . import utils
+from . import cards
 
 
 logger = logging.getLogger("krcg")
 
 
-class Deck(collections.Counter):
+class Deck(collections.Counter[cards.Card]):
     """A VTES deck, including meta information such as date, author and comments."""
 
-    def __init__(self, *args, **kwargs):
-        self.id = kwargs.pop("id", None)
-        self.author = kwargs.pop("author", None)
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
+        """Constructor."""
+        self.id: Optional[str] = kwargs.pop("id", None)
+        self.author: Optional[str] = kwargs.pop("author", None)
         super().__init__(*args, **kwargs)
-        self.event = None
-        self.place = None
-        self.date = None
-        self.tournament_format = None
-        self.players_count = 0
-        self.player = None
-        self.event_link = None
-        self.score = None
-        self.name = None
-        self.cards_comments = {}
-        self.comments = ""
+        self.event: Optional[str] = None
+        self.place: Optional[str] = None
+        self.date: Optional[datetime.date] = None
+        self.tournament_format: Optional[str] = None
+        self.players_count: int = 0
+        self.player: Optional[str] = None
+        self.event_link: Optional[str] = None
+        self.score: Optional[DeckScore] = None
+        self.name: Optional[str] = None
+        self.cards_comments: Dict[cards.Card, str] = {}
+        self.comments: str = ""
         # corner case: counting the number of "Raven" in "Camille Devereux" copies
-        self.raven = 0
-        self.parser = parser.Parser(self)
+        self.raven: int = 0
 
     @classmethod
     def from_txt(
         cls,
         input: TextIO,
-        id: str = None,
-        author: str = None,
+        id: Optional[str] = None,
+        author: Optional[str] = None,
         offset: int = 0,
         twda: bool = False,
-    ):
+        preface: bool = True,
+    ) -> "Deck":
         """Parse input and generate a Deck instance.
 
         Compatible with most text input formats (TWD, JOL, LackeyCCG, ...)
 
-        Set the `twda` arg to True when parsing TWDA entries, or it won't be perfect.
+        Set `twda` to True when parsing TWDA entries, or it won't be perfect.
+        Set `preface` to False when parsing pure card lists (e.g., no count).
 
-        Provide the offset when parsing a multi-deck stream for meaningful parsing logs.
+        Provide `offset` when parsing a multi-deck stream for meaningful parsing logs.
+
+        Args:
+            input: Text stream to parse.
+            id: Optional deck identifier.
+            author: Optional deck author.
+            offset: Line offset used for logging.
+            twda: Parse with TWDA-specific headers enabled.
+            preface: Expect a preface header section before cards.
+
+        Returns:
+            The parsed deck.
         """
+        from . import parser
+
         p = parser.Parser(cls(id=id, author=author))
-        p.parse(input, offset, twda)
+        p.parse(input, offset, twda, preface)
         return p.deck
 
     @classmethod
-    def from_amaranth(cls, uid: str):
+    def from_amaranth(cls, uid: str) -> "Deck":
         """Fetch a deck from Amaranth."""
-        r = requests.post("https://amaranth.vtes.co.nz/api/deck", data={"id": uid})
-        r.raise_for_status()
-        r = r.json()["result"]
-        ret = cls(id=uid, author=r.get("author", None))
-        ret.name = r.get("title", None)
-        ret.comments = r.get("description", "")
-        ret.date = arrow.get(r["modified"]).date()
+        response = requests.post(
+            "https://amaranth.vtes.co.nz/api/deck", data={"id": uid}
+        )
+        response.raise_for_status()
+        result: Dict[str, Any] = response.json()["result"]
+        ret = cls(id=uid, author=result.get("author", None))
+        ret.name = result.get("title", None)
+        ret.comments = result.get("description", "")
+        ret.date = arrow.get(result["modified"]).date()
         if not vtes.VTES:
             vtes.VTES.load()
-        for cid, count in r["cards"].items():
+        for cid, count in result["cards"].items():
             count = int(count)
             if count <= 0:
                 continue
@@ -82,23 +101,23 @@ class Deck(collections.Counter):
         return ret
 
     @classmethod
-    def from_vdb(cls, uid: str):
+    def from_vdb(cls, uid: str) -> "Deck":
         """Fetch a deck from VDB."""
-        r = requests.get("https://vdb.im/api/deck/" + uid)
-        r.raise_for_status()
-        r = r.json()
-        logger.debug("VDB replied: %s", r)
-        ret = cls(id=uid, author=r.get("author", r.get("owner", None)))
-        ret.name = r.get("name", None)
-        ret.comments = r.get("description", "")
+        response = requests.get("https://vdb.im/api/deck/" + uid)
+        response.raise_for_status()
+        data: Dict[str, Any] = response.json()
+        logger.debug("VDB replied: %s", data)
+        ret = cls(id=uid, author=data.get("author", data.get("owner", None)))
+        ret.name = data.get("name", None)
+        ret.comments = data.get("description", "")
         ret.date = (
-            email.utils.parsedate_to_datetime(r["timestamp"]).date()
-            if ("timestamp" in r)
+            email.utils.parsedate_to_datetime(data["timestamp"]).date()
+            if ("timestamp" in data)
             else None
         )
         if not vtes.VTES:
             vtes.VTES.load()
-        for cid, count in r["cards"].items():
+        for cid, count in data["cards"].items():
             count = int(count)
             if count <= 0:
                 continue
@@ -106,8 +125,30 @@ class Deck(collections.Counter):
         return ret
 
     @classmethod
-    def from_url(cls, url: str):
-        """Fetch from any deckbuilding website"""
+    def from_vtesdecks(cls, uid: str) -> "Deck":
+        """Fetch a deck from VTESDecks."""
+        response = requests.get("https://api.vtesdecks.com/1.0/decks/" + uid)
+        response.raise_for_status()
+        data: Dict[str, Any] = response.json()
+        logger.debug("VTESDecks replied: %s", data)
+        ret = cls(id=uid, author=data.get("author", data.get("owner", None)))
+        ret.name = data.get("name", None)
+        ret.comments = data.get("description", "")
+        date = data.get("modifyDate")
+        assert isinstance(date, str)
+        ret.date = datetime.datetime.fromisoformat(date).date()
+        if not vtes.VTES:
+            vtes.VTES.load()
+        for card in itertools.chain(data["crypt"], data["library"]):
+            count = int(card["number"])
+            if count <= 0:
+                continue
+            ret[vtes.VTES[int(card["id"])]] = count
+        return ret
+
+    @classmethod
+    def from_url(cls, url: str) -> "Deck":
+        """Fetch a deck from a supported deckbuilding website."""
         result = urllib.parse.urlparse(url)
         if result.netloc == "amaranth.vtes.co.nz":
             if result.fragment.startswith("deck/"):
@@ -121,14 +162,14 @@ class Deck(collections.Counter):
                 return cls.from_vdb(params["id"][0])
             elif result.fragment:
                 ret = cls()
-                ret.name = params.get("name", None)
-                ret.author = params.get("author", None)
-                ret.comments = params.get("description", "")
+                ret.name = params.get("name", [None])[0]
+                ret.author = params.get("author", [None])[0]
+                ret.comments = params.get("description", [""])[0]
                 if not vtes.VTES:
                     vtes.VTES.load()
                 for item in result.fragment.split(";"):
-                    card, count = item.split("=", 1)
-                    count = int(count)
+                    card, count_str = item.split("=", 1)
+                    count = int(count_str)
                     if count <= 0:
                         continue
                     ret[vtes.VTES[int(card)]] = count
@@ -136,15 +177,19 @@ class Deck(collections.Counter):
             elif result.path.startswith("/decks/"):
                 return cls.from_vdb(result.path[7:])
             raise ValueError("Unknown VDB URL format")
+        elif result.netloc == "vtesdecks.com":
+            if not result.path.startswith("/deck"):
+                raise ValueError("Unknown VTESDecks URL path")
+            return cls.from_vtesdecks(result.path[5:])
         raise ValueError("Unknown deck URL provider")
 
     def check(self) -> bool:
-        """Check a deck conforms to the rules IRT cards count
+        """Check a deck conforms to the rules regarding card counts.
 
-        Does not check grouping nor banished cards
+        Does not check grouping nor banned cards.
 
         Returns:
-            True if the deck checks out
+            True if the deck checks out.
         """
         res = True
         library_count = self.cards_count(lambda c: c.library)
@@ -186,7 +231,7 @@ class Deck(collections.Counter):
             "tournament_format": self.tournament_format,
             "players_count": self.players_count,
             "player": self.player,
-            "score": self.score,
+            "score": str(self.score) if self.score else None,
             "name": self.name,
             "author": self.author,
             "comments": self.comments,
@@ -206,16 +251,16 @@ class Deck(collections.Counter):
         }
 
         # form a section for each type with a header displaying the total
-        for kind, cards in self._sorted_library():
-            ret["library"]["cards"].append(
+        for kind, cards_ in self._sorted_library():
+            ret["library"]["cards"].append(  # type: ignore
                 {
                     "type": config.TYPE_ORDER[kind],
-                    "count": sum(count for _card, count in cards),
+                    "count": sum(count for _card, count in cards_),
                     "cards": [],
                 }
             )
-            for card, count in cards:
-                ret["library"]["cards"][-1]["cards"].append(
+            for card, count in cards_:
+                ret["library"]["cards"][-1]["cards"].append(  # type: ignore
                     {
                         "id": card.id,
                         "count": count,
@@ -226,6 +271,7 @@ class Deck(collections.Counter):
         return utils.json_pack(ret)
 
     def to_minimal_json(self) -> dict:
+        """Return a compact minimal JSON representation of the deck."""
         ret = {
             "id": self.id,
             "name": self.name,
@@ -233,20 +279,20 @@ class Deck(collections.Counter):
         }
         return utils.json_pack(ret)
 
-    def from_json(self, state) -> None:
-        """Initialize from a JSON dict generated by the `to_json()` method."""
+    def from_json(self, state: dict) -> None:
+        """Initialize from a JSON dict generated by `to_json()`."""
         self.id = state.get("id")
         self.event = state.get("event")
         self.place = state.get("place")
         if "date" in state:
             self.date = datetime.date.fromisoformat(state["date"])
         self.tournament_format = state.get("tournament_format")
-        self.players_count = state.get("players_count")
+        self.players_count = int(state.get("players_count", 0))
         self.player = state.get("player")
-        self.score = state.get("score")
+        self.score = DeckScore(state.get("score")) if state.get("score") else None
         self.name = state.get("name")
         self.author = state.get("author")
-        self.comments = state.get("comments")
+        self.comments = str(state.get("comments", ""))
         for card_id, count in state.get("cards", {}).items():
             c = vtes.VTES[int(card_id)]
             self[c] = count
@@ -262,30 +308,32 @@ class Deck(collections.Counter):
                 if card.get("comments"):
                     self.cards_comments[c] = card["comments"]
 
-    def __str__(self):
+    def __str__(self) -> str:
+        """Return the name of the deck, or "(No Name)" if it has no name."""
         return self.name or "(No Name)"
 
-    def __repr__(self):
+    def __repr__(self) -> str:
+        """Return the representation of the deck."""
         return f"<Deck #{self.id}: {self.name}>"
 
     @property
-    def crypt(self):
+    def crypt(self) -> list[tuple[cards.Card, int]]:
         """For convenience, list of crypt (card, count)."""
         return list(self.cards(lambda c: c.crypt))
 
     @property
-    def library(self):
+    def library(self) -> list[tuple[cards.Card, int]]:
         """For convenience, list of library (card, count)."""
         return list(self.cards(lambda c: c.library))
 
     def cards(self, condition: Optional[Callable] = None) -> Generator:
-        """Generator yielding (card_name, count), with an optional filter.
+        """Generator yielding (card, count), with an optional filter.
 
         Args:
-            condition: Condition each card must validate to be selected
+            condition: Condition each card must validate to be selected.
 
         Yields:
-            card_name, count
+            Pairs of (card, count).
         """
         for card, count in self.items():
             if condition and not condition(card):
@@ -296,10 +344,10 @@ class Deck(collections.Counter):
         """Generator yielding card names with an optional filter.
 
         Args:
-            condition: Condition each card must validate to be selected
+            condition: Condition each card must validate to be selected.
 
         Yields:
-            card_name
+            Card instance.
         """
         for card, _count in self.cards(condition):
             yield card
@@ -308,10 +356,10 @@ class Deck(collections.Counter):
         """Card counts with an optional filter.
 
         Args:
-            condition: Condition each card must validate to be selected
+            condition: Condition each card must validate to be selected.
 
         Returns:
-            The count of cards (matching the condition if any)
+            The count of cards (matching the condition if any).
         """
         return sum(count for _card, count in self.cards(condition))
 
@@ -319,10 +367,10 @@ class Deck(collections.Counter):
         """A generator that yields library cards sorted by type and name.
 
         Yields:
-            "type", [{card_name: count}]
+            Tuples of (type_index, list[(card, count)]).
         """
 
-        def _type_index(card_count):
+        def _type_index(card_count: tuple[cards.Card, int]) -> int:
             return config.TYPE_ORDER.index("/".join(sorted(card_count[0].types)))
 
         library_cards = sorted(
@@ -332,9 +380,9 @@ class Deck(collections.Counter):
                 utils.normalize(a[0].vekn_name),
             ),
         )
-        for kind, cards in itertools.groupby(library_cards, key=_type_index):
+        for kind, cards_ in itertools.groupby(library_cards, key=_type_index):
             # return a list so it can be iterated over multiple times
-            yield kind, list(cards)
+            yield kind, list(cards_)
 
     def to_txt(self, format: str = "twd") -> str:
         """Format to text. Default is TWD format."""
@@ -345,7 +393,7 @@ class Deck(collections.Counter):
         }.get(format, self._to_txt_twd)()
 
     def _to_txt_twd(self) -> str:
-        """A consistent deck display matching the most recent format used in TWDA.html
+        """A consistent deck display matching the most recent format used in TWDA.html.
 
         Cards are displayed in the order given by the config.TYPE_ORDER list.
         The output matches the newest TWDA entries format exactly.
@@ -382,7 +430,7 @@ class Deck(collections.Counter):
             lines.append("")
         cap = sorted(
             itertools.chain.from_iterable(
-                [card.capacity] * count for card, count in self.crypt
+                [card.capacity] * count for card, count in self.crypt if card.capacity
             )
         )
         cap_min = sum(cap[:4])
@@ -430,7 +478,7 @@ class Deck(collections.Counter):
                     ),
                     title=card.title.lower() if card.title else "",
                     clan=card.clans[0],
-                    group=card.group.upper(),
+                    group=(card.group or "").upper(),
                 )
             )
             if card in self.cards_comments:
@@ -438,24 +486,24 @@ class Deck(collections.Counter):
                 lines[-1] += f" -- {comment}"
         lines.append(f"\nLibrary ({self.cards_count(lambda c: c.library)} cards)")
         # form a section for each type with a header displaying the total
-        for i, (type_index, cards) in enumerate(self._sorted_library()):
+        for i, (type_index, cards_) in enumerate(self._sorted_library()):
             trifle_count = ""
             if type_index == 0:  # Master
                 trifle_count = sum(
                     count
-                    for card, count in cards
+                    for card, count in cards_
                     if card in vtes.VTES.search(bonus=["trifle"])
                 )
                 trifle_count = f"; {trifle_count} trifle" if trifle_count else ""
             cr = "\n" if i > 0 else ""
             lines.append(
                 f"{cr}{config.TYPE_ORDER[type_index]} "
-                f"({sum(count for _, count in cards)}{trifle_count})"
+                f"({sum(count for _, count in cards_)}{trifle_count})"
             )
-            for card, count in cards:
+            for card, count in cards_:
                 if card in self.cards_comments:
                     comment = self.cards_comments[card].replace("\n", " ").strip()
-                    lines.append(f"{count}x {card.vekn_name:<23}" f" -- {comment}")
+                    lines.append(f"{count}x {card.vekn_name:<23} -- {comment}")
                 else:
                     lines.append(f"{count}x {card.vekn_name}")
         return "\n".join(lines)
@@ -466,20 +514,20 @@ class Deck(collections.Counter):
         for card, count in self.crypt:
             lines.append(f"{count}x {card.vekn_name}")
         lines.append("")
-        for _, cards in self._sorted_library():
-            for card, count in cards:
+        for _, cards_ in self._sorted_library():
+            for card, count in cards_:
                 lines.append(f"{count}x {card.vekn_name}")
         return "\n".join(lines)
 
     def _to_txt_lackey(self) -> str:
         """Format used by LackeyCCG."""
 
-        def lackerize(name):
-            return unidecode.unidecode(card.vekn_name)
+        def lackerize(name: str) -> str:
+            return unidecode.unidecode(name)
 
         lines = []
-        for _, cards in self._sorted_library():
-            for card, count in cards:
+        for _, cards_ in self._sorted_library():
+            for card, count in cards_:
                 lines.append(f"{count}\t{lackerize(card.vekn_name)}")
         lines.append("Crypt:")
         for card, count in self.crypt:
@@ -487,18 +535,61 @@ class Deck(collections.Counter):
         return "\n".join(lines)
 
     def to_vdb(self) -> str:
-        """Generating vdb.smeaa.casa link to deck"""
-        link = "https://vdb.im/decks?"
-        link += urllib.parse.urlencode(
-            {
-                "name": self.name or "New KRCG Deck",
-                "author": self.author or self.player or "KRCG",
-            }
-        )
+        """Generate a vdb.im link to the deck."""
+        link = "https://vdb.im/decks/deck?"
+        params = {}
+        if self.name:
+            params["name"] = self.name
+        if self.author:
+            params["author"] = self.author
+        if params:
+            link += urllib.parse.urlencode(params)
         link += "#"
         for card, count in self.crypt:
             link += f"{card.id}={count};"
-        for _, cards in self._sorted_library():
-            for card, count in cards:
+        for _, cards_ in self._sorted_library():
+            for card, count in cards_:
                 link += f"{card.id}={count};"
-        return link[:-1]
+        return link[:-1]  # remove trailing semicolumn
+
+
+class DeckScore:
+    """Score: number of game wins (GW), round VP (VP), and finals VP (+X).
+
+    The score is a string like "2GW3+1", "1GW", "3VP", "1GW3", etc.
+    """
+
+    def __init__(self, s: Optional[str] = None):
+        """Constructor."""
+        self.game_wins = None
+        self.round_vps = None
+        self.finals_vps = None
+        if s is None:
+            return
+        score = re.match(
+            r"(\s*-?-?\s*(?P<game_wins>\d)\s*gw\s*"
+            r"(?P<round_vps>\d+(\.|,)?\d?)\s*(vp)?\s*)?"
+            r"(\s*-?-?\s*(?P<plus_mark>\+)?\s*(?P<finals_vps>\d(\.|,)?\d?)\s*"
+            r"(?(plus_mark).?|vp))?",
+            s.lower(),
+        )
+        if score is None or score.end() < 1:
+            raise ValueError("No score information")
+        self.game_wins = score.group("game_wins")
+        self.round_vps = score.group("round_vps")
+        if self.round_vps:
+            self.round_vps = self.round_vps.replace(",", ".")
+        self.finals_vps = score.group("finals_vps")
+        if self.finals_vps:
+            self.finals_vps = self.finals_vps.replace(",", ".")
+
+    def __str__(self) -> str:
+        """Return the compact string representation (e.g., 2GW3+1)."""
+        ret = ""
+        if self.game_wins:
+            ret += f"{self.game_wins}GW"
+        if self.round_vps:
+            ret += f"{self.round_vps}"
+        if self.finals_vps:
+            ret += f"+{self.finals_vps}"
+        return ret
