@@ -1,9 +1,10 @@
 """Deck list parser.
 
-It handles the legacy TWDA: many tricky formats used through this historic document.
-Only modifiy this file if you know what you're doing, and proceed with caution.
+Handles legacy TWDA formats and other deck list idiosyncrasies.
+Only modify this file if you know what you're doing, and proceed with caution.
 """
-from typing import TextIO, Tuple
+
+from typing import TextIO, Optional, MutableMapping, Any
 import datetime
 import enum
 import logging
@@ -13,6 +14,8 @@ import re
 import arrow
 
 from . import config
+from . import cards
+from . import deck
 from . import vtes
 from . import utils
 
@@ -24,8 +27,6 @@ _HEADERS = [
     "actin",
     "action",
     "acton",
-    "allies",
-    "ally",
     "average",
     "burn option",
     "capacity",
@@ -71,6 +72,11 @@ _HEADERS = [
     "vote",
     "and",
 ]
+# headers if first word only (do not match "political ally" as a header)
+_HEADERS_FIRST_WORD = [
+    "allies",
+    "ally",
+]
 _DISCIPLINES = {
     "abombwe",
     "animalism",
@@ -86,6 +92,7 @@ _DISCIPLINES = {
     "necromancy",
     "obeah",
     "obfuscate",
+    "oblivion",
     "obtenebration",
     "potence",
     "presence",
@@ -104,8 +111,15 @@ _DISCIPLINES = {
 _HEADERS_RE = (
     "(" + "|".join(re.escape(utils.normalize(h)) + "s?" for h in _HEADERS) + ")"
 )
+_HEADERS_FIRST_RE = (
+    "("
+    + "|".join(
+        re.escape(utils.normalize(h)) + "s?" for h in _HEADERS + _HEADERS_FIRST_WORD
+    )
+    + ")"
+)
 _HEADERS_RE = (
-    r"^(\s|/|-|\d|_)*{0}((\s|/|-)*{0})*".format(_HEADERS_RE)
+    r"^(\s|/|-|\d|_)*{0}((\s|/|-)*{1})*".format(_HEADERS_FIRST_RE, _HEADERS_RE)
     + r"(\s|\d|:|;|\.|\(|\)|\[|\]|/|-|=|,|"
     + r"cards?|carta|cars|total|min|max|avg|masters?|minions?|trifles?)*$"
 )
@@ -147,6 +161,7 @@ _DISCIPLINE_TRIGRAM = "|".join(
         "nec",
         "obe",
         "obf",
+        "obl",
         "obt",
         "pot",
         "pre",
@@ -197,6 +212,7 @@ _CLAN = "|".join(
         "blood brother",
         "toreador",
         "giovanni",
+        "hecata",
         "ahrimane",
         "akunanse",
         "brujah",
@@ -255,6 +271,8 @@ _PUNCTUATED_TRAIT = "|".join(
         "martyrdom",
         "pander",
         "presence",
+        # appears in name "touch of valeren"
+        "valeren",
         "vengeance",
         "vision",
         "mage",
@@ -265,7 +283,7 @@ _PUNCTUATED_TRAIT = "|".join(
         "giovanni",
         "goblin",
         "changeling",
-        # this one pften appears after double dashes in legacy deck lists:
+        # this one often appears after double dashes in legacy deck lists:
         # Bang Nakh -- Tiger's Claws
         # catch it here to avoid putting it in comments,
         # there's a matching alias in config.ALIASES
@@ -339,7 +357,6 @@ _NAKED_TRAIT = "|".join(
         "temporis",
         "thanatosis",
         "thaumaturgy",
-        "valeren",
         "vicissitude",
         "visceratika",
     ]
@@ -376,15 +393,29 @@ _RE = (
 
 
 class LineLogAdapter(logging.LoggerAdapter):
-    """Log line and deck"""
+    """Logger adapter that prefixes messages with line and deck IDs."""
 
-    def process(self, msg, kwargs):
+    def process(
+        self, msg: str, kwargs: MutableMapping[str, Any]
+    ) -> tuple[str, MutableMapping[str, Any]]:
+        """Process a message.
+
+        Args:
+            msg: The message to process.
+            kwargs: Additional keyword arguments.
+
+        Returns:
+            A tuple containing the processed message and the keyword arguments.
+        """
+        if not self.extra:
+            self.extra = {}
+        assert isinstance(self.extra, dict)
         self.extra.update(kwargs.get("extra", {}))
         return "[%6s][%s] %s" % (self.extra["line"], self.extra["deck"], msg), kwargs
 
 
 class Mark(enum.Enum):
-    """A comment Mark."""
+    """A comment mark type."""
 
     LINE = enum.auto()
     MULTILINE = enum.auto()
@@ -393,15 +424,27 @@ class Mark(enum.Enum):
 
 
 class Comment:
-    """Helper for comments parsing."""
+    """Helper class for comment parsing."""
 
-    def __init__(self, comment: str = "", card: object = None, mark: Mark = None):
+    def __init__(
+        self,
+        comment: str = "",
+        card: Optional[Any] = None,
+        mark: Optional[Mark] = None,
+    ):
+        """Constructor.
+
+        Args:
+            comment: The comment string.
+            card: The card associated with the comment.
+            mark: The mark type.
+        """
         self.card = card
         self.mark = mark
         self.string = comment
 
-    def __iadd__(self, comment: str):
-        """Add a comment line to self"""
+    def __iadd__(self, comment: str) -> "Comment":
+        """Add a comment line to the current comment string."""
         comment = comment.rstrip()
         if self.string:
             self.string += "\n"
@@ -412,10 +455,12 @@ class Comment:
         self.string += comment
         return self
 
-    def __bool__(self):
+    def __bool__(self) -> bool:
+        """Return True if the comment is not empty."""
         return bool(self.string)
 
-    def __str__(self):
+    def __str__(self) -> str:
+        """Return the comment string."""
         return self.string
 
     def finalize(self) -> None:
@@ -444,7 +489,7 @@ class Comment:
 
     @property
     def log(self) -> str:
-        """Log-friendly version of the comment (cropped at 83 chars)"""
+        """Return a log-friendly version of the comment (cropped at 83 chars)."""
         res = self.string.replace("\n", " ").strip()
         if len(res) > 83:
             res = res[:80] + "..."
@@ -452,39 +497,43 @@ class Comment:
 
     @property
     def multiline(self) -> bool:
-        """True if the comment is multiline"""
+        """Return True if the comment is multiline."""
         return len(self.string.split("\n")) > 1
 
 
 class Parser:
-    """Deck list parser. Holds the parsing context.
+    """Deck list parser. Holds the parsing context."""
 
-    Attributes:
-        deck (deck.Deck): the result, filled by the parsing operation
-        preface (bool): stays True until the beginning of the actual cards list
-        current_comment (Comment|None): current comment if any
-    """
-
-    def __init__(self, deck):
-        self.current_comment = None
-        self.preface = True
-        self.separator = False  # used only for additional checks on the TWDA
-        self.deck = deck
-        self.logger = logger
-
-    @property
-    def _previous_line(self):
-        return (getattr(self.logger, "extra", {}).get("line") or 1) - 1
-
-    def parse(self, input: TextIO, offset: int = 0, twda: bool = False) -> None:
-        """Parse given stream.
+    def __init__(self, d: deck.Deck) -> None:
+        """Constructor.
 
         Args:
-            offset: offset to add when parsing part of a bigger stream (for logs)
-            twda: if true, parse for TWDA headers
+            d: The deck to parse.
+        """
+        self.current_comment: Optional[Comment] = None
+        self.preface = True
+        self.separator = False  # used only for additional checks on the TWDA
+        self.deck: deck.Deck = d
+        self.logger: logging.Logger | logging.LoggerAdapter = logger
+
+    @property
+    def _previous_line(self) -> int:
+        return (getattr(self.logger, "extra", {}).get("line") or 1) - 1
+
+    def parse(
+        self, input: TextIO, offset: int = 0, twda: bool = False, preface: bool = True
+    ) -> None:
+        """Parse a deck list stream.
+
+        Args:
+            input: Text stream to parse.
+            offset: Offset to add when parsing part of a bigger stream (for logs).
+            twda: If True, parse TWDA headers.
+            preface: If True, expect a preface section before the cards.
         """
         if not vtes.VTES:
             vtes.VTES.load()
+        self.preface = preface
         for index, line in enumerate(input, 1):
             self.logger = LineLogAdapter(
                 logger, {"line": index + offset, "deck": self.deck.id}
@@ -497,21 +546,36 @@ class Parser:
         if not twda or self.deck.id not in config.TWDA_CHECK_DECK_FAILS:
             self.deck.check()
 
-    def parse_line(self, index: int, line: str, twda: bool):
+    def parse_line(self, index: int, line: str, twda: bool) -> None:
         """Parse a line of text."""
         # remove head/tail and misplaced spaces around punctuation for easy parsing
         line = line.rstrip()
         line = line.replace(" :", ":")
         line = line.replace("( ", "(")
         line = line.replace(" )", ")")
-        if twda and self.preface:
-            if self.parse_twda_headers(index, line):
+        if self.preface:
+            if twda and self.parse_twda_headers(index, line):
                 return
+            if self.parse_headers(index, line):
+                return
+        else:
+            # Always put a date, so if no date was parsed in headers,
+            # just put today as the date
+            if not self.deck.date:
+                self.deck.date = datetime.date.today()
+                if twda:
+                    self.logger.warning("No date found, using today")
+            # author is only set if different than player
+            if self.deck.author and not self.deck.player:
+                self.deck.player = self.deck.author
+                self.deck.author = None
+            if self.deck.author == self.deck.player:
+                self.deck.author = None
         card, count = self.get_card(line, twda)
         if card and count:
             self.deck.update({card: count})
 
-    def parse_twda_headers(self, index: int, line: str):
+    def parse_twda_headers(self, index: int, line: str) -> bool:
         """Parse a line of text for TWDA headers."""
         if index == 1:
             self.deck.event = line
@@ -532,13 +596,13 @@ class Parser:
             return True
         if not self.deck.tournament_format:
             try:
-                self.deck.tournament_format = re.match(r"^\s*(\d+R\+F)", line).group(1)
+                self.deck.tournament_format = re.match(r"\s*(\d+R\+F)", line).group(1)  # type: ignore
                 return True
             except AttributeError:
                 pass
         if not self.deck.players_count:
             try:
-                players_count = re.match(r"^\s*(\d+|\?+)\s*player", line).group(1)
+                players_count = re.match(r"\s*(\d+|\?+)\s*player", line).group(1)  # type: ignore
                 self.deck.players_count = int(players_count)
                 return True
             except AttributeError:
@@ -554,7 +618,7 @@ class Parser:
         # Newer lists provide an event link
         if not self.deck.event_link:
             try:
-                self.deck.event_link = re.match(r"^\s*(https?://.*)$", line).group(1)
+                self.deck.event_link = re.match(r"^\s*(https?://.*)$", line).group(1)  # type: ignore
                 return True
             except AttributeError:
                 pass
@@ -566,51 +630,69 @@ class Parser:
             if player:
                 self.deck.player = player
             return True
+        return False
+
+    def parse_headers(self, index: int, line: str) -> bool:
+        """Parse non-TWDA headers (name, author, score, etc.)."""
+        description = re.match(r"^\s*(D|d)escription\s*:?\s*", line)
+        if description:
+            line = line[description.end() :]
         if not self.deck.score:
             try:
-                score = re.match(
-                    r"-?-?\s*((?P<round_wins>\d)\s*(G|g)(W|w)\s*"
-                    r"(?P<round_vps>\d(\.|,)?\d?)\s*((v|V)(p|P))?\s*\+?\s*)?"
-                    r"(?P<final>\d(\.|,)?\d?)\s*(v|V)(p|P)",
-                    line,
-                )
-                if score.group("round_wins"):
-                    self.deck.score = "{}gw{} + {}vp in the final".format(
-                        score.group("round_wins"),
-                        score.group("round_vps"),
-                        score.group("final"),
-                    )
-                else:
-                    self.deck.score = "{}vp in the final".format(score.group("final"))
+                self.deck.score = deck.DeckScore(line)
                 return True
-            except AttributeError:
+            except (AttributeError, ValueError):
                 pass
         if not self.deck.name:
             try:
-                self.deck.name = re.match(
-                    r"^\s*((d|D)eck)?\s?(n|N)ame\s*:\s*(?P<name>.*)$", line
-                ).group("name")
+                self.deck.name = (
+                    re.match(r"^\s*((d|D)eck)?\s?(n|N)ame\s*:\s*(?P<name>.*)$", line)  # type: ignore
+                    .group("name")
+                    .strip()
+                )
                 return True
-            except AttributeError:
+            except (AttributeError, ValueError):
                 pass
         if not self.deck.author:
             try:
-                self.deck.author = re.match(
-                    r"(((c|C)reated|(d|D)eck)\s*(b|B)y|"
-                    r"(a|A)uthors?|(c|C)reators?)\s*(:|\s)\s*(?P<author>.*)$",
-                    line,
-                ).group("author")
+                self.deck.author = (
+                    re.match(
+                        r"\s*(((c|C)reated|(d|D)eck)\s*(b|B)y|"
+                        r"(a|A)uthors?|(c|C)reators?)\s*(:|\s)\s*(?P<author>.*)$",
+                        line,
+                    )
+                    .group("author")  # type: ignore
+                    .strip()
+                )
+                return True
+            except AttributeError:
+                pass
+        if not self.deck.player:
+            try:
+                self.deck.player = (
+                    re.match(
+                        r"\s*((p|P)layed\s*(b|B)y)|((p|P)layer)\s*(:|\s)\s*"
+                        r"(?P<player>.*)$",
+                        line,
+                    )
+                    .group("player")  # type: ignore
+                    .strip()
+                )
                 return True
             except AttributeError:
                 pass
-        # Always put a date, so if no date was parsed in headers,
-        # just put today as the date
         if not self.deck.date:
-            self.deck.date = datetime.date.today()
-            self.logger.warning("No date found, using today")
+            try:
+                self.deck.date = arrow.get(line, "MMMM Do YYYY").date()
+                return True
+            except arrow.parser.ParserMatchError:
+                pass
+        return False
 
-    def get_card(self, line: str, twda: bool = False) -> Tuple[object, int]:
-        """Try to find a card and count, register possible comment."""
+    def get_card(
+        self, line: str, twda: bool = False
+    ) -> tuple[Optional[cards.Card], int]:
+        """Try to find a card and count; register possible comment."""
         if re.match(_HEADERS_RE, utils.normalize(line)):
             return None, 0
         card, name, count, comment, mark = None, None, 0, "", None
@@ -637,20 +719,23 @@ class Parser:
                 if self.preface:
                     name = None
                 else:
-                    count = int(match.group("post_count") or 1)
-                    if not match.group("count_mark"):
+                    count = int(match.group("post_count") or 1)  # type: ignore
+                    if not match.group("count_mark"):  # type: ignore
                         # Be wary of disciplines: they are sometimes headers, but
-                        # distinguishing them from actual Master discipline cards
-                        # is not decidable so we log and ignore the line
+                        # indistinguishable from actual Master discipline cards
                         if name.strip(" :()[]-_*=") in _DISCIPLINES:
-                            self.logger.warning('improper discipline "%s"', line)
-                            return None, 0
+                            # if parsing TWDA, ignore the line and warn
+                            if twda:
+                                self.logger.warning('improper discipline "%s"', line)
+                                return None, 0
+                            # otherwise log as debug and count it as a discipline card
+                            self.logger.debug('naked discipline (no count) "%s"', line)
             # for evolutions (eg. Theo Bell (G6)) the name in TWDA
             # might not contain the group, we need to rely on the group in crypt tail
             if name and group and name[-1] != ")":
                 name += f" (g{group})"
             try:
-                card = vtes.VTES[name]
+                card = vtes.VTES[name]  # type: ignore
                 # special case for Camille / Raven to keep them distinct if the decklist
                 # predates the merge of the two crypt cards
                 if (
@@ -669,10 +754,10 @@ class Parser:
             self.logger.warning('discarded match "%s" inside comment "%s"', name, line)
             card, count = None, 0
         # do not match crypt tail expression on a library card
-        if card and match.group("crypt_tail") and not card.crypt:
+        if card and match.group("crypt_tail") and not card.crypt:  # type: ignore
             card, count = None, 0
         # do not match post count on a crypt card
-        if card and match.group("post_count") and card.crypt:
+        if card and match.group("post_count") and card.crypt:  # type: ignore
             card, count = None, 0
         # too many preface comments parse like cards in the TWDA
         if twda and name and not self.separator:
@@ -680,18 +765,18 @@ class Parser:
         # if no card was found, the whole line is a comment
         # if a card was found, a comment might still be present as a suffix
         if card:
-            comment = max(
-                match.span("parenthesis_comment"),
-                match.span("bracket_comment"),
-                match.span("line_comment"),
+            comment_span = max(
+                match.span("parenthesis_comment"),  # type: ignore
+                match.span("bracket_comment"),  # type: ignore
+                match.span("line_comment"),  # type: ignore
             )
             # extract the original comment, not the "normalized" parsed version
-            if comment > (-1, -1):
-                comment = line[comment[0] : comment[1]]
+            if comment_span > (-1, -1):
+                comment = line[comment_span[0] : comment_span[1]]
             else:
                 comment = ""
-            mark = match.group("comment_mark") or ""
-            if mark[:2] == "/*":
+            mark_match = match.group("comment_mark") or ""  # type: ignore
+            if mark_match[:2] == "/*":
                 if "*/" in line:
                     mark = Mark.LINE
                 else:
@@ -722,11 +807,13 @@ class Parser:
         self.preface = self.preface and not card
         return card, count
 
-    def comment(self, comment: str, card: object = None, mark: Mark = None) -> None:
-        """Handle a comment.
-
-        Log if we suspect the potential comment to be a parsing error
-        """
+    def comment(
+        self,
+        comment: str,
+        card: Optional[cards.Card] = None,
+        mark: Optional[Mark] = None,
+    ) -> None:
+        """Handle a comment and possibly log suspected parsing errors."""
         if not (comment or self.current_comment):
             return
         if comment:
@@ -790,7 +877,7 @@ class Parser:
         ):
             if mark == Mark.END:
                 self.current_comment += comment
-                comment = None
+                comment = ""
             self.current_comment.finalize()
             if self.current_comment:
                 current_card = self.current_comment.card
@@ -807,5 +894,5 @@ class Parser:
             self.current_comment = Comment(card=card, mark=mark)
         # append the parsed comment, even if it is a blank line
         if comment or self.current_comment:
-            self.current_comment += comment
+            self.current_comment += comment  # type: ignore
         # do nothing on a blank line if we don't have a current comment
